
/*

    File: main.c
    
    Copyright (C) 1999,2000,2003,2004,2006,2007 by Wolfgang Zekoll <wzk@quietsche-entchen.de>

    This source is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This source is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <stdarg.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/socket.h>
#include <signal.h>
#include <netdb.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/time.h>
#include <pwd.h>

#include "lib.h"
#include "ip-lib.h"
#include "procinfo.h"
#include "smtp.h"


char	*program =		"";
char	progname[80] =		"";

int	debug =			0;
int	alsosyslog =		0;
int	supervised =		0;
int	checkconfig =		0;

char	bindarg[80] =		"";
int	daemonmode =		0;


tokentab_t smmodetab[] = {
	{ "prim-sec",		SM_PRIMSEC },
	{ "round-robin",	SM_ROUNDROBIN },
	{ "",			0 }
    };

int printerror(int rc, char *type, char *format, ...)
{
	char	tag[30], error[800];
	va_list	ap;

	va_start(ap, format);
	vsnprintf (error, sizeof(error) - 2, format, ap);
	va_end(ap);
	noctrl(error);

	*tag = 0;
	if (*type != 0)
		snprintf (tag, sizeof(tag) - 2, "%s: ", type);

	if (debug != 0)
		fprintf (stderr, "%s: %s%s\n", program, tag, error);

	if (debug == 0  ||  alsosyslog != 0)
		syslog(LOG_NOTICE, "%s%s", tag, error);

	if (rc == EXIT_ZEROSTATUS)
		exit (0);

	if (rc != 0)
		exit (rc);

	return (0);
}

void signalhandler(int sig)
{
	if (sig == SIGPIPE) {

		/* Ignore here and make write() return -1. */

		signal(SIGPIPE, signalhandler);
		return;
		}

	printerror(0, "+INFO", "caught signal #%d", sig);
	exit(0);
}

int checkload(config_t *config, char *loadpar, char *current, int size)
{
	int	i;
	char	*p, *r, line[300];
	double	par, val;
	FILE	*fp;

	if (*loadpar == 0)
		return (0);

	if ((fp = fopen(UPTIME_FILENAME, "r")) == NULL) {
		printerror(0, "-ERR", "can't open uptimefile %s, error= %s", UPTIME_FILENAME, strerror(errno));
		return (550);
		}

	fgets(line, sizeof(line), fp);
	fclose (fp);

	p = skip_ws(noctrl(line));
	copy_string(current, p, size);

	r = loadpar;
	for (i=1; i<=3; i++) {
		if (*(r = skip_ws(r)) == 0)
			break;

		par = strtod(r, &r);
		val = strtod(p, &p);
		if (val < par  &&  par > 0)	/* `10 0 2' means ignore second value */
			return (0);		/* limit not reached */
		}

	return (421);
}

int writestatfile(smtp_t *x, char *status)
{
	if (getstatfp() == NULL)
		return (0);

	if (pi.statfp != NULL) {
		ftruncate(fileno(pi.statfp), 0);
		rewind(pi.statfp);
		fprintf (pi.statfp, "%s %s %u %lu %s:%u %s %s %s:%u %s:%u %s\n",
				PROXYNAME, program, getpid(),
				x->started,
				x->interface.ipnum, x->interface.port,
				/* x->client.ipnum */ x->ipnum, /* x->client.name */ x->client,
				x->s.ipnum, x->s.port,
				x->origdst.ipnum, x->origdst.port,
				status);
		fflush(pi.statfp);
		}

	return (0);
}


int printconfig(config_t *config)
{
	fprintf (stderr, "%s: printconfig() disabled\n", program);
	exit (1);
}

  /* ----------------------------------------------------------------
   *                       D I S A B L E D

#define	printyesno(x, y, z)	(y + z != 0? printf ("%s\t %s\n", x, (y == 0)? "no": "yes"): 0)
#define printnum(x, y)		((y > 0)? printf ("%s\t %u\n", x, y): 0)
#define printoct(x, y)		((y > 0)? printf ("%s\t %03o\n", x, y): 0)
#define	printstring(x, y)	((y != NULL  &&  *y != 0)? printf("%s\t %s\n", x, y): 0)

int printconfig(config_t *config)
{
	char	*p;

	if (*config->interface != 0) {
		printstring ("interface", config->interface);
		printyesno ("have interfaces", config->haveinterfaces, 1);
		printyesno ("valid interface", config->subconfig, 1);
		printf ("\n");
		}

	printyesno ("debug", debug, 1);
	printyesno ("daemonmode", daemonmode, 0);
	printstring ("username", config->user.name);
	printstring ("acp", config->acp);
	printyesno ("allow", config->allowdeny.permitted, 1);
	printstring ("ccp", config->ccp);
	printstring ("clientdir", config->clientdir);
	printyesno ("droppath", config->droppath, 1);
	printyesno ("etrn", config->etrn, 1);
	printyesno ("supervised", supervised, 1);
	printstring ("# port", ((daemonmode != 0)? bindarg: ""));
	printstring ("pidfile",  config->pidfile);

	if (isempty(&config->rcpt) != 0)
		printstring ("rcptlist", noctrl(config->rcpt.string));

	printstring ("server", config->server);
	printstring ("senderlist", config->senderlist);
	printstring ("source-ip", config->srcip);
	printnum ("timeout", config->timeout);
	printstring ("varname", config->varname);

	printyesno ("clamav-scan", config->scanmail, 1);
	printstring ("clamav-quarantine", config->clamav.qd.dir);
	printyesno ("clamav-passonerror", config->clamav.passonerror, 1);
	printyesno ("clamav-passonvirus", config->clamav.passonvirus, 1);
	p = gettmpdir();  printstring ("clamav-tmpdir", p);

	printyesno ("spamc-scan", config->spamd.scan, 1);
	printstring ("spamc-tag", config->spamd.spamtag);
	printstring ("spamc-cmd", config->spamd.cmd);
	printyesno ("spamc-quarantine", config->spamd.quarantine, 1);
	printstring ("# spamc-qdir", config->spamd.qd.dir);

	printyesno ("rbl-logonly", config->rbl.logonly, 1);

	return (0);
}

     ---------------------------------------------------------------- */


static int get_yesno(char **from, char *par, char *filename, int lineno)
{
	char	word[80];

	if (**from == 0) {
		fprintf (stderr, "%s: missing parameter: %s, %s:%d\n", program, par, filename, lineno);
		exit (1);
		}

	get_word(from, word, sizeof(word));
	if (strcmp(word, "yes") == 0)
		return (1);
	else if (strcmp(word, "no") == 0)
		return (0);

	fprintf (stderr, "%s: bad parameter value: %s, parameter= %s, %s:%d\n", program, word, par, filename, lineno);
	exit (1);
	
	return (0);
}

static char *get_parameter(char **from, char *par, char *value, int size,
		char *filename, int lineno)
{
	if (**from == 0) {
		fprintf (stderr, "%s: missing parameter: %s, %s:%d\n", program, par, filename, lineno);
		exit (1);
		}

	copy_string(value, *from, size);
	if (strcmp(value, "-") == 0)
		*value = 0;

	return (value);
}

static long get_number(char **from, char *par, char *filename, int lineno)
{
	long	num;
	char	*p, word[80];

	if (**from == 0) {
		fprintf (stderr, "%s: missing parameter: %s, %s:%d\n", program, par, filename, lineno);
		exit (1);
		}

	get_word(from, word, sizeof(word));
	*from = skip_ws(*from);
	if (**from != 0) {
		fprintf (stderr, "%s: bad number: %s, %s:%d\n", program, par, filename, lineno);
		exit (1);
		}

	num = strtol(word, &p, 10);
	if (*p != 0) {
		fprintf (stderr, "%s: not a number: %s, %s:%d\n", program, word, filename, lineno);
		exit (1);
		}

	return (num);
}


int setword(char *word, char *name, tokentab_t *tab, char *filename, int lineno)
{
	int	i;

	strlwr(word);
	for (i=1; tab[i].word[0] != 0; i++) {
		if (strcmp(word, tab[i].word) == 0)
			return (tab[i].opc);
		}

	printerror(1, "-ERR", "unknown %s parameter: %s:%d", name, filename, lineno);
	return (1);
}

int setlistpar(config_t *config, string_t *s, char *list)
{
	char	*p, word[200];

	p = list;
	while (*(p = skip_ws(p)) != 0  &&  *get_quoted(&p, ',', word, sizeof(word)) != 0) {
		noctrl(word);
		addline(s, word, " ");
		}

	return (0);
}

int setsmtpserver(config_t *config, char *list)
{
	char	*p, server[SMTP_SERVERLEN];

	p = list;
	while (*(p = skip_ws(p)) != 0) {
		get_quoted(&p, ',', server, sizeof(server));

		if (config->server.count >= MAX_SMTPSERVER)
			printerror(1, "-ERR", "to many SMTP servers: %s", server);

		copy_string(config->server.server[config->server.count++], server, SMTP_SERVERLEN);
		}

	return (0);
}


int readconfig(config_t *config, char *filename, char *interface, char *ipnum)
{
	int	lineno, insection, sectioncount, havesection;
	char	*p, word[80], line[2000];
	FILE	*fp;

	if ((fp = fopen(filename, "r")) == NULL) {
		fprintf (stderr, "%s: can't open configuration file: %s\n", program, filename);
		exit (1);
		}

	/*
	 * Normalize parameters.
	 */

	if (interface == NULL)
		interface = "";

	if (ipnum == NULL)
		ipnum = "";


	sectioncount = havesection = 0;
	if (*interface == 0)
		insection = 1;
	else
		insection = 0;


	lineno = 0;
	while (fgets(line, sizeof(line), fp) != NULL) {
		lineno++;
		if ((p = strchr(line, '#')) != NULL)
			*p = 0;

		p = skip_ws(noctrl(line));
		if (*p == 0)
			continue;


		if (*line == '[') {
			if (checkconfig != 0  &&  debug != 0)
				fprintf (stderr, ":: %s\n", line);

			/*
			 * The beginning of an interface and/or client ip
			 * specific configuration section.
			 */

			if (*interface == 0) {

				/*
				 * If the interface parameter is not set, we
				 * read only global configuration parameters,
				 * no sections.
				 */

				break;
				}

			insection = 0;
			if ((p = strchr(line, ']')) != NULL)
				*p = 0;

			p = skip_ws(&line[1]);


			/*
			 * Get the first word which is (the optional)
			 * interface name.
			 */

			get_word(&p, word, sizeof(word));
			if (*word == 0) {
				printerror(0 | ERR_STDERR, "-ERR", "configuration file error: %s:%d, line= %s",
						filename, lineno, line);
				exit (1);
				}

			/*
			 * Compare section interface.
			 */

			if (strcmp(word, "-") == 0  ||  strcmp(word, "*") == 0)
				insection = 1;
			else {
				/*
				 * Notice that we have interface specific
				 * sections.
				 */

				config->haveinterfaces = 1;
				sectioncount++;

				if (strcmp(interface, "-") == 0  ||  strcmp(interface, "*") == 0)
					insection = 1;		/* parameter wildcard */
				else if (strcmp(word, interface) == 0)
					insection = 1;

				if (insection == 1)
					havesection = 1;
				}

			if (1)
				/* disable the following IP check - 2007-01-16/wzk */ ;
			else if (insection == 1  &&  *(p = skip_ws(p)) != 0) {

				/*
				 * Check the client's IP number.
				 */

				if (ipnum == 0) {

					/*
					 * No ip number given.  This can usually
					 * only happen when we check the configuration
					 * with `-F'.  In this case ignore any client
					 * specific settings.
					 */

					insection = 0;
					}

				get_word(&p, word, sizeof(word));
				if (strcmp(word, ipnum) != 0)
					insection = 0;
				}

			continue;
			}


		if (insection == 0)
			continue;	/* The current section is not for us. */

		if (checkconfig != 0  &&  debug != 0)
			fprintf (stderr, ":: %s\n", line);

		get_word(&p, word, sizeof(word));
		strlwr(word);
		p = skip_ws(p);

		if (*p == '=') {
			p++;
			p = skip_ws(p);
			}

		if (strcmp(word, "debug") == 0)
			debug = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "debugcontrol") == 0  ||  strcmp(word, "debug-control") == 0)
			get_parameter(&p, word, config->debug.controlfile, sizeof(config->debug.controlfile), filename, lineno);

		else if (strcmp(word, "supervised") == 0)
			supervised = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "daemonmode") == 0)
			daemonmode = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "username") == 0)
			get_parameter(&p, word, config->user.name, sizeof(config->user.name), filename, lineno);
		else if (strcmp(word, "timeout") == 0) {
			config->timeout = get_number(&p, word, filename, lineno);
			if (config->timeout <= 0)
				config->timeout = 60;
			}

		else if (strcmp(word, "statdir") == 0)
			get_parameter(&p, word, statdir, sizeof(statdir), filename, lineno);
		else if (strcmp(word, "maxloadavg") == 0)
			get_parameter(&p, word, config->loadavg, sizeof(config->loadavg), filename, lineno);

/*		else if (strcmp(word, "interface") == 0  ||  strcmp(word, "interfaces") == 0)
 *			get_parameter(&p, word, config->interface, sizeof(config->interface), filename, lineno);
 */

		else if (strcmp(word, "server") == 0) {
			char	server[SMTP_SERVERLEN];

/*			if (config->server.count >= config->server.max) {
 *				config->server.max += 8;
 *				config->server.server = reallocate(config->server.server, config->server.max * SMTP_SERVERLEN);
 *				}
 */

			get_parameter(&p, word, server, sizeof(server), filename, lineno);
			setsmtpserver(config, server);
			}
		else if (strcmp(word, "server-type") == 0) {
			char	*r, typepar[300];

			get_parameter(&p, word, typepar, sizeof(typepar), filename, lineno);
			if ((r = strchr(typepar, ' ')) != NULL)
				*r++ = 0;

			config->server.mode = setword(typepar, word, smmodetab, filename, lineno);
			if (r != NULL)
				copy_string(config->server.laststartfile, r, sizeof(config->server.laststartfile));
			}

		else if (strcmp(word, "droppath") == 0)
			config->droppath = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "etrn") == 0)
			config->etrn = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "pidfile") == 0)
			get_parameter(&p, word, config->pidfile, sizeof(config->pidfile), filename, lineno);
		else if (strcmp(word, "rcptlist") == 0) {
			char	par[400];

			get_parameter(&p, word, par, sizeof(par), filename, lineno);
			setlistpar(config, &config->rcpt, par);
			}
		else if (strcmp(word, "senderlist") == 0) {
			char	par[400];

			get_parameter(&p, word, par, sizeof(par), filename, lineno);
			setlistpar(config, &config->senderlist, par);
			}
		else if (strcmp(word, "srcip") == 0  ||  strcmp(word, "source-ip") == 0)
			get_parameter(&p, word, config->srcip, sizeof(config->srcip), filename, lineno);
		else if (strcmp(word, "acp") == 0)
			get_parameter(&p, word, config->acp, sizeof(config->acp), filename, lineno);
		else if (strcmp(word, "ccp") == 0)
			get_parameter(&p, word, config->ccp, sizeof(config->ccp), filename, lineno);
		else if (strcmp(word, "clientip-dir") == 0)
			get_parameter(&p, word, config->clientdir, sizeof(config->clientdir), filename, lineno);
		else if (strcmp(word, "setenv") == 0  ||  strcmp(word, "varname") == 0)
			get_parameter(&p, word, config->varname, sizeof(config->varname), filename, lineno);

		else if (strcmp(word, "allow") == 0)
			config->allowdeny.permitted = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "deny") == 0)
			config->allowdeny.permitted = 1 - get_yesno(&p, word, filename, lineno);

		else if (strcmp(word, "tmpdir") == 0  ||  strcmp(word, "clamav-tmpdir") == 0) {
			char	dir[300];

			get_parameter(&p, word, dir, sizeof(dir), filename, lineno);
			settmpdir(dir);
			}

		else if (strcmp(word, "clamav-scan") == 0)
			config->scanmail = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "clamav-qdir") == 0  ||  strcmp(word, "clamav-quarantine") == 0)
			get_parameter(&p, word, config->clamav.qd.dir, sizeof(config->clamav.qd.dir), filename, lineno);
		else if (strcmp(word, "clamav-dirmode") == 0) {
			char	mode[80];

			get_parameter(&p, word, mode, sizeof(mode), filename, lineno);
			config->clamav.qd.dirmode = setword(mode, word, dirmodetab, filename, lineno);
			}
		else if (strcmp(word, "clamav-index") == 0) {
			char	mode[80];

			get_parameter(&p, word, mode, sizeof(mode), filename, lineno);
			config->clamav.qd.indexstyle = setword(mode, word, indexmodetab, filename, lineno);
			}
		else if (strcmp(word, "clamav-passonerror") == 0)
			config->clamav.passonerror = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "clamav-passonvirus") == 0)
			config->clamav.passonvirus = get_yesno(&p, word, filename, lineno);


		else if (strcmp(word, "spamc-scan") == 0)
			config->spamd.scan = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "spamc-tag") == 0)
			get_parameter(&p, word, config->spamd.spamtag, sizeof(config->spamd.spamtag), filename, lineno);
		else if (strcmp(word, "spamc-cmd") == 0)
			get_parameter(&p, word, config->spamd.cmd, sizeof(config->spamd.cmd), filename, lineno);
		else if (strcmp(word, "spamc-quarantine") == 0)
			config->spamd.quarantine = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "spamc-qdir") == 0) {
			get_parameter(&p, word, config->spamd.qd.dir, sizeof(config->spamd.qd.dir), filename, lineno);
			config->spamd.quarantine = 1;
			}
		else if (strcmp(word, "spamc-dirmode") == 0) {
			char	mode[80];

			get_parameter(&p, word, mode, sizeof(mode), filename, lineno);
			config->spamd.qd.dirmode = setword(mode, word, dirmodetab, filename, lineno);
			}
		else if (strcmp(word, "spamc-treshhold") == 0)
			config->spamd.treshhold = get_number(&p, word, filename, lineno);
		else if (strcmp(word, "spamc-index") == 0) {
			char	mode[80];

			get_parameter(&p, word, mode, sizeof(mode), filename, lineno);
			config->spamd.qd.indexstyle = setword(mode, word, indexmodetab, filename, lineno);
			}

		else if (strcmp(word, "rbl") == 0  ||  strcmp(word, "rbl-server") == 0) {
			char	list[1024];

			get_parameter(&p, word, list, sizeof(list), filename, lineno);
			setlistpar(config, &config->rbl.list, list);
			}
		else if (strcmp(word, "rbl-mode") == 0) {
			char	mode[80];

			get_parameter(&p, word, mode, sizeof(mode), filename, lineno);
			config->rbl.mode = setword(mode, word, rblmodetab, filename, lineno);
			}
		else if (strcmp(word, "rbl-firstmatch") == 0)
			config->rbl.usefirst = get_yesno(&p, word, filename, lineno);
		else if (strcmp(word, "rbl-logonly") == 0)
			config->rbl.mode = RBL_LOGONLY;

		else {
			fprintf (stderr, "%s: unknown parameter: %s, %s:%d\n", program, word, filename, lineno);
			exit (1);
			}
		}

	fclose (fp);
	config->subconfig = (*interface == 0  ||  havesection != 0)? 1: 0;
	if (sectioncount == 0)
		havesection = 1;

	return (havesection);
}

int setuserid(config_t *config)
{
	char    *p;
	struct passwd *pw;

	if (*config->user.name == 0)
		strcpy(config->user.name, "nobody");

	if (*(p = config->user.name) >= '0'  &&  *p <= '9') {
		config->user.uid = strtoul(config->user.name, &p, 10);
		config->user.gid = 65534;
		if (*p == '.') {
			p++;
			config->user.gid = strtoul(p, &p, 10);
			}

		if (*p != 0)
			printerror(1, "-PROXY", "invalid username: username= %s", config->user.name);
		}
	else if ((pw = getpwnam(config->user.name)) != NULL) {
		config->user.uid = pw->pw_uid;
		config->user.gid = pw->pw_gid;
		}
	else
		printerror(1, "-PROXY", "no such user: username= %s", config->user.name);

	if (*(p = getpidfile()) != 0) {
		if (chown(p, config->user.uid, config->user.gid) != 0) {
			printerror(0, "-INFO", "can't change pidfile permissions, pidfile= %s, error= %s",
					p, strerror(errno));
			}
		}

	if (setregid(config->user.gid, config->user.gid) != 0)
		printerror(1, "-PROXY", "can't set gid: username= %s", config->user.name);
	else if (setreuid(config->user.uid, config->user.uid) != 0)
		printerror(1, "-PROXY", "can't set uid: username= %s", config->user.name);

	return (0);
}


void missing_arg(int c, char *string)
{
	fprintf (stderr, "%s: missing arg: -%c, %s\n", program, c, string);
	exit (-1);
}

int main(int argc, char *argv[])
{
	int	c, i, k;
	char	*p, option[80];
	config_t config;
	

	if ((p = strrchr(argv[0], '/')) == NULL)
		program = argv[0];
	else {
		copy_string(progname, &p[1], sizeof(progname));
		program = progname;
		}

	memset(&config, 0, sizeof(config_t));
	config.timeout    = SMTP_TIMEOUT;
	config.accepttime = SMTP_ACCEPTTIME;
	copy_string(config.varname, "SMTP_", sizeof(config.varname));
	openlog(program, LOG_PID, LOG_MAIL);

	copy_string(config.loadavg, "10 0 0", sizeof(config.loadavg));
	config.clamav.statuscode    = 250;
	config.clamav.qd.dirmode    = 1;

	config.spamd.quarantine     = 0;
	config.spamd.statuscode     = 250;
	config.clamav.qd.dirmode    = 1;

	config.allowdeny.permitted  = 1;	/* allow connections */

	config.earlytalker.timeout  = 1000;	/* 1000 milliseconds */
	config.earlytalker.rc       = 0;	/* 421 is the default */
	config.earlytalker.delay    = 0;	/* do not wait for "MAIL FROM" before disconnect */

	copy_string(config.spamd.spamtag, SPAM_TAG, sizeof(config.spamd.spamtag));
	copy_string(config.spamd.cmd, SPAM_SCANCMD, sizeof(config.spamd.cmd));

	config.server.mode          = SM_ROUNDROBIN;
	copy_string(config.server.laststartfile, "/var/tmp/smtp.last", sizeof(config.server.laststartfile));

	k = 1;
	while (k < argc  &&  argv[k][0] == '-'  &&  argv[k][1] != 0) {
		copy_string(option, argv[k++], sizeof(option));
		for (i=1; (c = option[i]) != 0; i++) {
			if (c == 'd')
				debug++;
			else if (c == 'a') {
				if (k >= argc)
					missing_arg(c, "acp");

				copy_string(config.acp, argv[k++], sizeof(config.acp));
				}
			else if (c == 'b') {
				if (k >= argc)
					missing_arg(c, "rbl server");

				setlistpar(&config, &config.rbl.list, argv[k++]);
				}
			else if (c == 'c') {
				if (k >= argc)
					missing_arg(c, "ccp");

				copy_string(config.ccp, argv[k++], sizeof(config.ccp));
				}
			else if (c == 'e')
				config.etrn = 1;
			else if (c == 'f'  ||  c == 'F') {
				if (c == 'F')
					checkconfig = 1;

				if (k >= argc)
					missing_arg(c, "configuration file");

				copy_string(config.configfile, argv[k++], sizeof(config.configfile));
				readconfig(&config, config.configfile, "", "");
				}
			else if (c == 'l') {
				if (k >= argc)
					missing_arg(c, "client log directory");

				copy_string(config.clientdir, argv[k++], sizeof(config.clientdir));
				}
			else if (c == 'm') {
				if (k >= argc)
					missing_arg(c, "masq domain");

				copy_string(config.senderdomain, argv[k++], sizeof(config.senderdomain));
				}
			else if (c == 'o') {
				if (k >= argc)
					missing_arg(c, "statdir");

				copy_string(statdir, argv[k++], sizeof(statdir));
				}
			else if (c == 'p') {
				if (k >= argc)
					missing_arg(c, "pidfile");

				copy_string(config.pidfile, argv[k++], sizeof(config.pidfile));
				}
			else if (c == 'q') {
				if (k >= argc)
					missing_arg(c, "source ip");

				copy_string(config.srcip, argv[k++], sizeof(config.srcip));
				}
			else if (c == 'r') {
				if (k >= argc)
					missing_arg(c, "domain list");

				setlistpar(&config, &config.rcpt, argv[k++]);
				}
			else if (c == 's') {
				if (k >= argc)
					missing_arg(c, "domain list");

				setlistpar(&config, &config.senderlist, argv[k++]);
				}
			else if (c == 'S')
				supervised = 1;
			else if (c == 't') {
				if (k >= argc)
					missing_arg(c, "timeout");

				config.timeout = atoi(argv[k++]);
				if (config.timeout < 1)
					config.timeout = 60;
				}
			else if (c == 'u') {
				if (config.droppath == 0)
					config.droppath = 1;
				else
					config.droppath = 2;
				}
			else if (c == 'y') {
				if (config.spamd.scan != 0) {
					if (k >= argc)
						missing_arg(c, "spam client command");

					copy_string(config.spamd.cmd, argv[k++], sizeof(config.spamd.cmd));
					}

				config.spamd.scan = 1;
				}
			else if (c == 'z') {
				if (config.scanmail != 0) {
					if (k >= argc)
						missing_arg(c, "quarantine folder");

					copy_string(config.clamav.qd.dir, argv[k++], sizeof(config.clamav.qd.dir));
					}

				config.scanmail = 1;
				}
			else if (c == 'Z') {
				if (config.clamav.passonerror == 0)
					config.clamav.passonerror = 1;
				else if (config.clamav.passonvirus == 0)
					config.clamav.passonvirus = 1;
				}
			else if (c == 'D') {
				if (daemonmode == 0)
					daemonmode = 1;
				else {
					if (k >= argc)
						missing_arg(c, "port");

					copy_string(bindarg, argv[k++], sizeof(bindarg));
					}
				}
			else if (c == 'T') {
				if (k >= argc)
					missing_arg(c, "temporary directory");

				settmpdir(argv[k++]);
				}
			else if (c == 'U') {
				if (k >= argc)
					missing_arg(c, "username");

				copy_string(config.user.name, argv[k++], sizeof(config.user.name));
				}
			else if (c == 'V') {
				printf ("%s %s\n", program, VERSION);
				exit (0);
				}
			else {
				if (isatty(2))
					fprintf (stderr, "%s: unknown option: -%c\n", program, c);

				exit (1);
				}
			}
		}


	/*
	 * Changed the server parameter logic: if an argument is given
	 * on the command line, use is it as (default) upstream server.
	 * If no argument is there and we have a configuration file
	 * we go with it and expect an interface specific configuration
	 * section to define the server. 2007-01-16/wzk
	 */

	if (config.server.count > 0)
		/* We got the server from the configuration file. */ ;
	else if (k < argc) {
		if (*argv[k] == '/') {
			config.argv = &argv[k];
			k = argc;
			}
		else {
			while (k < argc)
				setsmtpserver(&config, argv[k++]);
			}
		}
	else if (*config.configfile != 0)
		/* move on */ ;
	else {
   		if (isatty(2))
   			fprintf (stderr, "%s: missing server argument\n", program);

		exit (1);
		}


	/*
	 * checkconfig has been deactivated for now.
	 */

	if (checkconfig != 0) {
		char	*interface, *ipnum;

		if (k < argc) {
			interface = argv[k++];
			ipnum = (k < argc)? argv[k++]: "";

			readconfig(&config, config.configfile, interface, ipnum);
			}

		printconfig(&config);
		exit (0);
		}


	if (config.scanmail != 0) {
		if (*config.clamav.qd.dir != 0) {
			createdir(config.clamav.qd.dir);
 			}
		}

	if (config.spamd.scan != 0) {
		if (config.spamd.quarantine != 0) {
			if (*config.spamd.qd.dir == 0)
				copy_string(config.spamd.qd.dir, config.clamav.qd.dir, sizeof(config.spamd.qd.dir));

			if (*config.spamd.qd.dir != 0)
				createdir(config.spamd.qd.dir);
			}
		}

	/*
	 * Check for additional arguments.
	 */

	if (k < argc) {
		if (isatty(2))
			fprintf (stderr, "%s: unexpected extra arguments\n", program);

		exit (1);
		}




	signal(SIGHUP, signalhandler);
	signal(SIGINT, signalhandler);
	signal(SIGQUIT, signalhandler);
	signal(SIGUSR1, signalhandler);
	signal(SIGUSR2, signalhandler);
	signal(SIGTERM, signalhandler);
	signal(SIGPIPE, signalhandler);
	atexit(exithandler);

	if (daemonmode != 0) {
		unsigned int port;
		int	sock, pid;
		char	interface[80];

		*interface = 0;
		p = bindarg;
		if (strchr(p, ':') != NULL)
			get_quoted(&p, ':', interface, sizeof(interface));

		port = strtoul(p, NULL, 10);
		if (port == 0)
			port = 25;

		signal(SIGCHLD, SIG_IGN);
		sock = bind_to_port(interface, port);

		if (debug != 0  ||  supervised != 0)
			;
		else if ((pid = fork()) > 0)
			exit (0);

		setpidfile(config.pidfile);
		if (getuid() == 0)
			setuserid(&config);

		syslog(LOG_NOTICE, "+INFO: smtp.proxy started in daemon mode, user= %s", config.user.name);
		acceptloop(sock);
		signal(SIGCHLD, SIG_DFL);
		}


	/*
	 * Here we are always have a client session.
	 */

	setnumvar("PID", getpid());
	proxy_request(&config);

	exit (0);
}


