
/*

    File: rbl.c

    Copyright (C) 2007  by Wolfgang Zekoll <wzk@quietsche-entchen.de>
 
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>

#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include "lib.h"
#include "ip-lib.h"
#include "smtp.h"


tokentab_t rblmodetab[] = {
	{ "reject",	RBL_REJECT },
	{ "quarantine",	RBL_QUARANTINE },
	{ "log-only",	RBL_LOGONLY },
	{ "logonly",	RBL_LOGONLY },
	{ "delay",	RBL_DELAYED },
	{ "delayed",	RBL_DELAYED },
	{ "",		0 }
    };


typedef struct {
    char	arec[300];
    char	txtrec[300];
    } lookup_t;

char *reverseip(char *ip, char *rev, int len)
{
	int	i;
	char	*p, num[4][20];

	p = ip;
	for (i=0; i < 4; i++) {
		if (*p == 0) {
			printerror(0, "-ERR", "IP number error: %s", ip);
			return (NULL);
			}

		get_quoted(&p, '.', num[i], sizeof(num[i]));
		}

	if (*p != 0) {
		printerror(0, "-ERR", "IP number error: %s", ip);
		return (NULL);
		}

	snprintf (rev, len - 2, "%s.%s.%s.%s", num[3], num[2], num[1], num[0]);
	return (rev);
}

int checkblacklist(smtp_t *x, char *name, char *result, int size)
{
	int	c, i, k, n, rc, len;
	char	buffer[520];

	rc = 0;
	*result = 0;
	size -= 2;

	/* get the A record */
	res_init();
	if ((len = res_query(name, C_IN, T_A, buffer, sizeof(buffer))) < 0)
		return (0);		/* no match */
	else if (len > 0) {
/*
 * Answer from RBL server is uninteresting, all that counts is taht
 * there is one.
 * 
 *		if (len > sizeof(buffer) - 2)
 *			len = sizeof(buffer) - 2;
 *
 *		buffer[len] = 0;
 */
		rc = 1;			/* positive match */
		}

	/* try to get the TXT record */
/*	res_init(); */
	if ((len = res_query(name, C_IN, T_TXT, buffer, sizeof(buffer))) <= 0)
		return (rc);		/* return what we have */
	else {
		if (len > sizeof(buffer) - 2)
			len = sizeof(buffer) - 2;

		buffer[len] = 0;
		rc = 1;
		}

	i = sizeof(HEADER);
	while (i < len  &&  (n = buffer[i]) != 0) {	/* skip question */
		i += 1 + n;
		}


	i += 1 + NS_INT16SZ + NS_INT32SZ;
	i += NS_INT16SZ + NS_INT16SZ + NS_INT32SZ;	/* TYPE, CLASS and TTL */

	n = (buffer[i] << 8) | buffer[i+1];		/* length of complete answer block */
	i += 2;
	if (n > size - 2)
		n = size - 2;

	/*
	 * n is changed to the length of only the first RR in the
	 * answer block, any second or further answer RR is
	 * ignored.
	 */

	n = buffer[i++];
	if (n > size - 2)
		n = size - 2;

	for (k=0; k < n; k++) {
		if ((c = buffer[i++]) <= 32)
			c = ' ';

		result[k] = c;
		}

	result[k] = 0;
	noctrl(result);

	return (rc);
}

int checkrbl(smtp_t *x, char *ipnum, char *msg, int size)
{
	int	count;
	char	*p, rip[100], server[200], name[200], result[100];

	if (reverseip(ipnum, rip, sizeof(rip)) == NULL)
		return (0);	/* accept due to error */

	p = x->config->rbl.list.string;
	count = 0;
	while (*get_quoted(&p, ' ', server, sizeof(server)) != 0) {
		snprintf (name, sizeof(name) - 2, "%s.%s", rip, server);
		if (checkblacklist(x, name, result, sizeof(result)) != 0) {
			printerror(0, "+INFO", "rbl match, server= %s, ipnum= %s, TXT= %s%s",
					server, ipnum, result,
					(x->config->rbl.mode == RBL_LOGONLY)? ", logonly= yes": "");
			count++;
			if (count == 1) {
				if (*result != 0)
					snprintf (msg, size - 2, "server= %s, info= %s", server, result);
				else
					snprintf (msg, size - 2, "server= %s", server);
				}

			if (x->config->rbl.usefirst != 0)
				break;
			}
		}

	if (count == 0) {
		if (x->config->rbl.verbose != 0)
			printerror(0, "+INFO", "client not blacklisted, ip= %s", ipnum);
		}

	if (x->config->rbl.mode == RBL_LOGONLY)
		count = 0;

	return (count);
}

